# -*- coding: utf-8 -*-
# Автор: Гусев Илья
# Описание: Класс рифм.

from typing import Tuple
from rupo.main.markup import Word
from rupo.util.preprocess import VOWELS


class Rhymes(object):
    """
    Поиск рифм.
    """

    @staticmethod
    def is_rhyme(word1: Word, word2: Word, score_border: int=4, syllable_number_border: int=4) -> bool:
        """
        Проверка рифмованности 2 слов.

        :param word1: первое слово для проверки рифмы, уже акцентуированное (Word).
        :param word2: второе слово для проверки рифмы, уже акцентуированное (Word).
        :param score_border: граница определния рифмы, чем выше, тем строже совпадение.
        :param syllable_number_border: ограничение на номер слога с конца, на который падает ударение.
        :return result: является рифмой или нет.
        """
        features1 = Rhymes.__get_rhyme_profile(word1)
        features2 = Rhymes.__get_rhyme_profile(word2)
        count_equality = 0
        for i in range(len(features1[1])):
            for j in range(i, len(features2[1])):
                if features1[1][i] == features2[1][j]:
                    if features1[1][i] in VOWELS:
                        count_equality += 3
                    else:
                        count_equality += 1
        if features1[2] == features2[2] and features1[2] != '' and features2[2] != '':
            count_equality += 2
        elif features1[3] == features2[3] and features1[3] != '' and features2[3] != '':
            count_equality += 1
        return features1[0] == features2[0] and count_equality >= score_border and \
               features1[0] <= syllable_number_border

    @staticmethod
    def __get_rhyme_profile(word: Word) -> Tuple[int, str, str, str]:
        """
        Получение профиля рифмовки (набора признаков для сопоставления).

        :param word: уже акцентуированное слово (Word).
        :return profile: профиль рифмовки.
        """
        # TODO: Переход на фонетическое слово, больше признаков.
        syllable_number = 0
        accented_syllable = ''
        next_syllable = ''
        next_char = ''
        syllables = list(reversed(word.syllables))
        for i in range(len(syllables)):
            syllable = syllables[i]
            if syllable.accent != -1:
                if i != 0:
                    next_syllable = syllables[i - 1].text
                accented_syllable = syllables[i].text
                if syllable.accent + 1 < len(word.text):
                    next_char = word.text[syllable.accent + 1]
                syllable_number = i
                break
        return syllable_number, accented_syllable, next_syllable, next_char



# -*- coding: utf-8 -*-
# Автор: Гусев Илья
# Описание: Модуль марковских цепей.

import os
import pickle
from collections import Counter
from typing import List

import numpy as np

from rupo.main.markup import Markup
from rupo.util.vocabulary import Vocabulary
from rupo.files.reader import Reader, FileTypeEnum


class MarkovModelContainer(object):
    """
    Марковские цепи.
    """
    def __init__(self, dump_filename: str, vocabulary: Vocabulary, markup_dump_path: str=None):
        self.transitions = list()
        self.vocabulary = vocabulary
        self.dump_filename = dump_filename

        # Делаем дамп модели для ускорения загрузки.
        if os.path.isfile(dump_filename):
            self.load()
        else:
            for i in range(len(self.vocabulary.words)):
                self.transitions.append(Counter())
            i = 0
            markups = Reader.read_markups(markup_dump_path, FileTypeEnum.XML, is_processed=True)
            for markup in markups:
                self.add_markup(markup)
                i += 1
                if i % 500 == 0:
                    print(i)
            self.save()

    def save(self):
        with open(self.dump_filename, "wb") as f:
            pickle.dump(self.transitions, f, pickle.HIGHEST_PROTOCOL)
        self.vocabulary.save()

    def load(self):
        with open(self.dump_filename, "rb") as f:
            self.transitions = pickle.load(f)
        self.vocabulary.load()

    def generate_chain(self, words: List[int]) -> List[Counter]:
        """
        Генерация переходов в марковских цепях с учётом частотности.

        :param words: вершины цепи.
        :return: обновленные переходы.
        """
        for i in range(len(words) - 1):
            current_word = words[i]
            next_word = words[i+1]
            self.transitions[current_word][next_word] += 1
        return self.transitions

    def add_markup(self, markup: Markup) -> None:
        """
        Дополнение цепей на основе разметки.

        :param markup: разметка.
        """
        words = []
        for line in markup.lines:
            for word in line.words:
                try:
                    self.vocabulary.get_word_index(word)
                    words.append(self.vocabulary.get_word_index(word))
                except IndexError:
                    print("Слово не из словаря.")
                    pass

        # Генерируем переходы.
        self.generate_chain(list(reversed(words)))

    def get_model(self, word_indices: List[int]) -> np.array:
        """
        Получение языковой модели.

        :param word_indices: индексы предыдущих слов.
        :return: языковая модель (распределение вероятностей для следующего слова).
        """
        l = len(self.transitions)
        if len(word_indices) == 0 or len(self.transitions[word_indices[-1]]) == 0:
            model = np.full(len(self.transitions), 1/l, dtype=np.float)
        else:
            transition = self.transitions[word_indices[-1]]
            s = sum(transition.values())
            model = np.zeros(len(self.transitions), dtype=np.float)
            for index, p in transition.items():
                model[index] = p/s
        return model



# -*- coding: utf-8 -*-
# Автор: Гусев Илья
# Описание: Тесты к классификатору метра.

import unittest
import jsonpickle

from rupo.main.markup import Markup
from rupo.accents.dict import AccentDict
from rupo.main.phonetics import Phonetics
from rupo.metre.metre_classifier import MetreClassifier, ClassificationResult, AccentCorrection


class TestMetreClassifier(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.accent_dict = AccentDict()

    def test_classification_result(self):
        result = ClassificationResult(5)
        result.additions["iambos"].append(AccentCorrection(0, 0, 0, "", 0))
        self.assertEqual(result, jsonpickle.decode(result.to_json()))

    def test_metre_classifier(self):
        text = "Горит восток зарёю новой.\n" \
               "Уж на равнине, по холмам\n" \
               "Грохочут пушки. Дым багровый\n" \
               "Кругами всходит к небесам."
        markup, result = MetreClassifier.improve_markup(Phonetics.process_text(text, self.accent_dict))
        self.assertIsInstance(markup, Markup)
        self.assertIsInstance(result, ClassificationResult)
        self.assertEqual(result.get_metre_errors_count(), 0)
        self.assertEqual(result.metre, "iambos")

        text = "Буря мглою небо кроет,\n" \
               "Вихри снежные крутя;\n" \
               "То, как зверь, она завоет,\n" \
               "То заплачет, как дитя..."
        markup, result = MetreClassifier.improve_markup(Phonetics.process_text(text, self.accent_dict))
        self.assertEqual(result.get_metre_errors_count(), 0)
        self.assertEqual(result.metre, "choreios")

        text = "Буря мглою небо парад,\n" \
               "Вихри снежные крутя;\n" \
               "То, как зверь, она завоет,\n" \
               "То заплачет, как дитя..."
        markup, result = MetreClassifier.improve_markup(Phonetics.process_text(text, self.accent_dict))
        self.assertEqual(result.get_metre_errors_count(), 1)
        self.assertEqual(result.metre, "choreios")

        text = "На стеклах нарастает лед,\n"\
               "Часы твердят: «Не трусь!»\n"\
               "Услышать, что ко мне идет,\n"\
               "И мертвой я боюсь.\n"\
               "Как идола, молю я дверь;\n"\
               "«Не пропускай беду!»\n"\
               "Кто воет за стеной, как зверь,\n"\
               "Кто прячется в саду?"
        markup, result = MetreClassifier.improve_markup(Phonetics.process_text(text, self.accent_dict))
        self.assertEqual(result.get_metre_errors_count(), 0)
        self.assertEqual(result.metre, "iambos")

        text = "Вот уж вечер. Роса\n" \
               "Блестит на крапиве.\n"\
               "Я стою у дороги,\n"\
               "Прислонившись к иве.\n"\
               "От луны свет большой\n"\
               "Прямо на нашу крышу.\n"\
               "Где-то песнь соловья\n"\
               "Хорошо и тепло,\n"\
               "Как зимой у печки.\n"\
               "И березы стоят,\n"\
               "Как большие свечки.\n"\
               "И вдали за рекой,\n"\
               "Видно, за опушкой,\n"\
               "Сонный сторож стучит\n"\
               "Мертвой колотушкой."
        markup, result = MetreClassifier.improve_markup(Phonetics.process_text(text, self.accent_dict))
        self.assertEqual(result.get_metre_errors_count(), 0)
        self.assertTrue(result.metre == "dolnik3" or result.metre == "dolnik2")

        text = "Глыбу кварца разбили молотом,\n" \
               "И, веселым огнем горя,\n" \
               "Заблестели крупинки золота\n" \
               "В свете тусклого фонаря.\n" \
               "И вокруг собрались откатчики:\n" \
               "Редкий случай, чтоб так, в руде!\n" \
               "И от ламп заплясали зайчики,\n" \
               "Отражаясь в черной воде...\n" \
               "Прислонившись к мокрой стене,\n" \
               "Мы стояли вокруг.\n" \
               "Курили,\n" \
               "Прислонившись к мокрой стене,\n" \
               "И мечтательно говорили\n" \
               "Не о золоте — о весне.\n" \
               "И о том, что скоро, наверно,\n" \
               "На заливе вспотеет лед\n" \
               "И, снега огласив сиреной,\n" \
               "Наконец придет пароход...\n" \
               "Покурили еще немного,\n" \
               "Золотинки в кисет смели\n" \
               "И опять — по своим дорогам,\n" \
               "К вагонеткам своим пошли.\n" \
               "Что нам золото? В дни тяжелые\n" \
               "Я от жадности злой не слеп.\n" \
               "Самородки большие, желтые\n" \
               "Отдавал за табак и хлеб.\n" \
               "Не о золоте были мысли...\n" \
               "В ночь таежную у костра\n" \
               "Есть над чем поразмыслить в жизни,\n" \
               "Кроме\n" \
               "Золота-серебра."
        markup, result = MetreClassifier.improve_markup(Phonetics.process_text(text, self.accent_dict))
        self.assertEqual(result.get_metre_errors_count(), 0)
        self.assertTrue(result.metre == "dolnik3" or result.metre == "dolnik2")



# -*- coding: utf-8 -*-
# Автор: Гусев Илья
# Описание: Тесты для словаря ударений.

import unittest
import os

from rupo.accents.dict import AccentDict
from rupo.util.preprocess import VOWELS
from rupo.settings import DICT_TXT_PATH, DICT_TRIE_PATH


class TestAccentDict(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.dict = AccentDict()

    def test_load_and_create(self):
        self.assertTrue(os.path.exists(DICT_TXT_PATH))
        self.assertTrue(os.path.exists(DICT_TRIE_PATH))
        os.remove(DICT_TRIE_PATH)
        AccentDict()
        self.assertTrue(os.path.exists(DICT_TRIE_PATH))

    def test_get_accents(self):
        self.assertCountEqual(self.dict.get_accents("данный", AccentDict.AccentType.PRIMARY), [1])
        self.assertCountEqual(self.dict.get_accents("союза", AccentDict.AccentType.PRIMARY), [2])
        self.assertCountEqual(self.dict.get_accents("англосакс", AccentDict.AccentType.SECONDARY), [0])
        self.assertCountEqual(self.dict.get_accents("англосакс", AccentDict.AccentType.ANY), [0, 6])
        self.assertCountEqual(self.dict.get_accents("пора", AccentDict.AccentType.PRIMARY), [1, 3])

    def test_accent_only_in_vowels(self):
        for word, accents in self.dict.get_all():
            for accent in accents:
                self.assertIn(word[accent[0]], VOWELS)




# -*- coding: utf-8 -*-
# Автор: Гусев Илья
# Описание: Шаблоны ритма.

from collections import defaultdict
from typing import List, Dict
from enum import Enum


class FlagEnum(Enum):
    ZERO_OR_ONE = "?"
    ZERO_TO_INF = "*"
    ONE = ""


class Group:
    """
    Группа - выражение в круглых скобках и её модификатор.
    """
    def __init__(self, exp: str, flag: FlagEnum) -> None:
        """
        :param exp: выражение в круглых скобках.
        :param flag: модификатор.
        """
        self.exp = exp
        self.flag = flag
        self.extended = []  # type: List[str]

    def __str__(self):
        return self.exp + str(self.flag) + ": " + str(self.extended)

    def __repr__(self):
        return self.__str__()


class Patterns:
    @staticmethod
    def compile_pattern(metre_pattern: str, l: int) -> List[str]:
        """
        Скомпилировать заданное выражение.

        :param metre_pattern: выражение.
        :param l: итоговое количество символов.
        :return: скомпилированные шаблоны.
        """
        group = Group(metre_pattern, FlagEnum.ONE)
        Patterns.__process_group(group, l)
        return list(set([st for st in group.extended if len(st) == l]))

    @staticmethod
    def __process_group(group: Group, l: int) -> None:
        """
        Обработка группы. Записывает в extended результаты.

        :param group: группа.
        :param l: максимальное количество символов.
        """
        # Высчитываем группы следующего уровня.
        children_groups = Patterns.__find_groups(group.exp)
        if len(children_groups) == 0:
            # Нижний уровень.
            if group.flag == FlagEnum.ZERO_OR_ONE or group.flag == FlagEnum.ONE:
                group.extended = [group.exp]
            elif group.flag == FlagEnum.ZERO_TO_INF:
                group.extended = [group.exp * i for i in range(1, l // len(group.exp) + 1)]
            else:
                assert False
        else:
            # Спускаемся.
            for child in children_groups:
                Patterns.__process_group(child, l)
            # Поднимаемся, параллельно проставляем extended
            group.extended = Patterns.__next(children_groups, 0, "", l)

    @staticmethod
    def __find_groups(expression: str) -> List[Group]:
        """
        Высчитываем группы выражения.

        :param expression: выражение.
        :return: группы.
        """
        groups = []
        counter = 0
        begin = -1
        for i in range(len(expression)):
            if expression[i] == "(":
                counter += 1
                if counter == 1:
                    begin = i + 1
            if expression[i] == ")":
                if counter == 1:
                    flag = FlagEnum.ONE  # type: FlagEnum
                    if i + 1 < len(expression) and (expression[i + 1] == "?" or expression[i + 1] == "*"):
                        flag = FlagEnum(expression[i + 1])
                    groups.append(Group(expression[begin:i], flag))
                counter -= 1
            assert counter >= 0
        assert counter == 0
        return groups

    @staticmethod
    def __next(groups: List[Group], index: int, pattern: str, l: int) -> List[str]:
        """
        Получение всех возможных шаблонов на заданных группах заданной длины.

        :param groups: группы.
        :param index: индекс текущей группы в массиве.
        :param pattern: текущий набранный шаблон
        :param l: максимальное количество символов.
        :return: все возможные шаблоны.
        """
        # База рекурсии.
        if len(pattern) > l or index >= len(groups):
            if sum(groups[i].flag == FlagEnum.ONE for i in range(index, len(groups))) == 0:
                return [pattern, ]
            return []
        result = []
        flag = groups[index].flag
        if flag == FlagEnum.ZERO_TO_INF:
            # Продолжаем все варианты *
            for i in range(len(groups[index].extended)):
                result += Patterns.__next(groups, index, pattern + groups[index].extended[i], l)
            # Пропускаем *
            result += Patterns.__next(groups, index + 1, pattern, l)
        elif flag == FlagEnum.ZERO_OR_ONE:
            # Пропускаем ?
            result += Patterns.__next(groups, index + 1, pattern, l)
            # Продолжаем ?
            for st in groups[index].extended:
                result += Patterns.__next(groups, index + 1, pattern + st, l)
        elif flag == FlagEnum.ONE:
            for st in groups[index].extended:
                result += Patterns.__next(groups, index + 1, pattern + st, l)
        return result


class CompiledPatterns(object):
    """
    Скомпилированные шаблоны (нужно, чтобы не пересчитывать их каждый раз).
    """
    def __init__(self, metres: Dict[str, str], border: int) -> None:
        self.compilations = defaultdict(lambda: defaultdict(lambda: ""))
        for metre_name, metre_pattern in metres.items():
            for i in range(1, border+1):
                self.compilations[metre_name][i] = Patterns.compile_pattern(metre_pattern, i)

    def get_patterns(self, metre_name: str, syllables_count: int) -> List[str]:
        """
        Получить или посчитать шаблоны метра для заданного метра.

        :param metre_name: название метра.
        :param syllables_count: количество слогов в шаблоне.
        :return: итоговые скомпилированные шаблоны метра.
        """
        return self.compilations[metre_name][syllables_count]



# -*- coding: utf-8 -*-
# Автор: Гусев Илья
# Описание: Набор внешних методов для работы с библиотекой.

from typing import List, Tuple

from rupo.main.phonetics import Phonetics
from rupo.main.markup import Markup
from rupo.accents.dict import AccentDict
from rupo.accents.classifier import MLAccentClassifier
from rupo.metre.metre_classifier import MetreClassifier, ClassificationResult
from rupo.files.reader import FileTypeEnum, Reader
from rupo.files.writer import Writer
from rupo.rhymes.rhymes import Rhymes
from rupo.generate.markov import MarkovModelContainer
from rupo.generate.generator import Generator
from rupo.util.vocabulary import Vocabulary


class Global:
    """
    Глобальные ресурсы.
    """
    accent_dict = None  # type: AccentDict
    accent_classifier = None  # type: MLAccentClassifier
    vocabulary = None  # type: Vocabulary
    markov = None  # type: MarkovModelContainer
    generator = None  # type: Generator

    @classmethod
    def get_dict(cls) -> AccentDict:
        if cls.accent_dict is None:
            cls.accent_dict = AccentDict()
        return cls.accent_dict

    @classmethod
    def get_classifier(cls) -> MLAccentClassifier:
        if cls.accent_classifier is None:
            cls.accent_classifier = MLAccentClassifier(cls.get_dict())
        return cls.accent_classifier

    @classmethod
    def get_vocabulary(cls, dump_path: str, markup_path: str) -> Vocabulary:
        if cls.vocabulary is None:
            cls.vocabulary = Vocabulary(dump_path, markup_path)
        return cls.vocabulary

    @classmethod
    def get_markov(cls, dump_path: str, vocab_dump_path: str, markup_path: str) -> MarkovModelContainer:
        if cls.markov is None:
            vocab = cls.get_vocabulary(vocab_dump_path, markup_path)
            cls.markov = MarkovModelContainer(dump_path, vocab, markup_path)
        return cls.markov

    @classmethod
    def get_generator(cls, dump_path: str, vocab_dump_path: str, markup_path: str) -> Generator:
        if cls.generator is None:
            cls.generator = Generator(cls.get_markov(dump_path, vocab_dump_path, markup_path),
                                      cls.get_vocabulary(vocab_dump_path, markup_path))
        return cls.generator


def get_accent(word: str) -> int:
    """
    :param word: слово.
    :return: ударение слова.
    """
    return Phonetics.get_improved_word_accent(word, Global.get_dict(), Global.get_classifier())


def get_word_syllables(word: str) -> List[str]:
    """
    :param word: слово.
    :return: его слоги.
    """
    return [syllable.text for syllable in Phonetics.get_word_syllables(word)]


def count_syllables(word: str) -> int:
    """
    :param word: слово.
    :return: количество слогов в нём.
    """
    return len(Phonetics.get_word_syllables(word))


def get_markup(text: str) -> Markup:
    """
    :param text: текст.
    :return: его разметка по словарю.
    """
    return Phonetics.process_text(text, Global.get_dict())


def get_improved_markup(text: str) -> Tuple[Markup, ClassificationResult]:
    """
    :param text: текст.
    :return: его разметка по словарю, классификатору метру и  ML классификатору.
    """
    markup = Phonetics.process_text(text, Global.get_dict())
    return MetreClassifier.improve_markup(markup, Global.get_classifier())


def classify_metre(text: str) -> str:
    """
    :param text: текст.
    :return: его метр.
    """
    return MetreClassifier.classify_metre(Phonetics.process_text(text, Global.get_dict())).metre


def generate_markups(input_path: str, input_type: FileTypeEnum, output_path: str, output_type: FileTypeEnum) -> None:
    """
    Генерация разметок по текстам.

    :param input_path: путь к папке/файлу с текстом.
    :param input_type: тип файлов с текстов.
    :param output_path: путь к файлу с итоговыми разметками.
    :param output_type: тип итогового файла.
    """
    markups = Reader.read_markups(input_path, input_type, False, Global.get_dict(), Global.get_classifier())
    writer = Writer(output_type, output_path)
    writer.open()
    for markup in markups:
        writer.write_markup(markup)
    writer.close()


def is_rhyme(word1: str, word2: str) -> bool:
    """
    :param word1: первое слово.
    :param word2: второе слово.
    :return: рифмуются ли слова.
    """
    markup_word1 = get_markup(word1).lines[0].words[0]
    markup_word1.set_accents([get_accent(word1)])
    markup_word2 = get_markup(word2).lines[0].words[0]
    markup_word2.set_accents([get_accent(word2)])
    return Rhymes.is_rhyme(markup_word1, markup_word2)


def generate_poem(markup_path: str, dump_path: str, vocab_dump_path: str, metre_schema: str="-+",
                  rhyme_pattern: str="abab", n_syllables: int=8) -> str:
    """
    Сгенерировать стих по данным из разметок.

    :param markup_path: путь к разметкам.
    :param dump_path: путь, куда сохранять модель.
    :param vocab_dump_path: путь, куда сохранять словарь.
    :param metre_schema: схема метра.
    :param rhyme_pattern: схема рифм.
    :param n_syllables: количество слогов в строке.
    :return: стих.
    """
    generator = Global.get_generator(dump_path, vocab_dump_path, markup_path)
    return generator.generate_poem(metre_schema, rhyme_pattern, n_syllables)


def generate_poem_by_line(markup_path: str, dump_path: str, vocab_dump_path: str,
                          line, rhyme_pattern="abab") -> str:
    """
    Сгенерировать стих по первой строчке.

    :param markup_path: путь к разметкам.
    :param dump_path: путь, куда сохраняется модель.
    :param vocab_dump_path: путь, куда сохраняется словарь.
    :param line: первая строчка
    :param rhyme_pattern: схема рифм.
    :return: стих.
    """
    generator = Global.get_generator(dump_path, vocab_dump_path, markup_path)
    return generator.generate_poem_by_line(line, rhyme_pattern, Global.get_dict(), Global.get_classifier())


def get_word_rhymes(word: str, vocab_dump_path: str, markup_path: str=None) -> List[str]:
    """
    Поиск рифмы для данного слова.

    :param word: слово.
    :param vocab_dump_path: путь, куда сохраняется словарь.
    :param markup_path: путь к разметкам.
    :return: список рифм.
    """
    markup_word = get_markup(word).lines[0].words[0]
    markup_word.set_accents([get_accent(word)])
    rhymes = []
    vocabulary = Global.get_vocabulary(vocab_dump_path, markup_path)
    for i in range(len(vocabulary.words)):
        if Rhymes.is_rhyme(markup_word, vocabulary.get_word(i)):
            rhymes.append(vocabulary.get_word(i).text.lower())
    return rhymes



# -*- coding: utf-8 -*-
# Автор: Гусев Илья
# Описание: Тесты для модуля рифм.

import unittest

from rupo.main.markup import Syllable, Word
from rupo.rhymes.rhymes import Rhymes


class TestRhymes(unittest.TestCase):
    def test_rhyme(self):
        self.assertTrue(Rhymes.is_rhyme(Word(0, 4, "тишь", [Syllable(0, 4, 0, "тишь", 1)]),
                                        Word(0, 8, "грустишь", [Syllable(0, 3, 0, "гру"),
                                                                Syllable(3, 8, 1, "стишь", 5)])))
        self.assertFalse(Rhymes.is_rhyme(Word(0, 8, "наизусть", [Syllable(0, 2, 0, "на"), Syllable(2, 4, 1, "из"),
                                                                 Syllable(4, 8, 2, "усть", 4)]),
                                         Word(0, 6, "сестра", [Syllable(0, 3, 0, "сест"),
                                                               Syllable(3, 6, 1, "ра", 5)])))


# -*- coding: utf-8 -*-
# Автор: Гусев Илья
# Описание: Тесты марковских цепей.

import unittest
import os

from rupo.settings import EXAMPLES_DIR, MARKUP_XML_EXAMPLE
from rupo.util.vocabulary import Vocabulary
from rupo.generate.markov import MarkovModelContainer


class TestMarkov(unittest.TestCase):
    def test_markov(self):
        vocab_dump_file = os.path.join(EXAMPLES_DIR, "vocab.pickle")
        markov_dump_file = os.path.join(EXAMPLES_DIR, "markov.pickle")
        vocabulary = Vocabulary(vocab_dump_file, MARKUP_XML_EXAMPLE)
        markov = MarkovModelContainer(markov_dump_file, vocabulary, MARKUP_XML_EXAMPLE)
        self.assertTrue(os.path.exists(vocab_dump_file))
        self.assertTrue(os.path.exists(markov_dump_file))
        os.remove(vocab_dump_file)
        os.remove(markov_dump_file)
        self.assertEqual(len(vocabulary.words), len(markov.transitions))
        self.assertEqual(sum([sum(transition.values()) for transition in markov.transitions]), len(vocabulary.words)-1)


# -*- coding: utf-8 -*-
# Автор: Гусев Илья
# Описание: Тесты для модуля классификации ударений.

import unittest

from rupo.accents.classifier import MLAccentClassifier
from rupo.accents.dict import AccentDict


class TestAccentClassifier(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.accent_dict = AccentDict()
        cls.accent_classifier = MLAccentClassifier(cls.accent_dict)

    def test_accent_classifier(self):
        self.assertEqual(len([0, 2, 2, 1, 2, 2, 2, 1, 2]),
                         len([self.accent_classifier.classify_accent(word) for word in
                             ["волки", "перелив", "карачун", "пипярка", "пепелац", "гиппогриф",
                              "хвосторог", "стартап", "квинтолап"]]))



# -*- coding: utf-8 -*-
# Автор: Гусев Илья
# Описание: Модуль создания стихотворений.

import numpy as np
from numpy.random import choice
from typing import List

from rupo.generate.filters import Filter, MetreFilter, RhymeFilter
from rupo.main.phonetics import Phonetics
from rupo.metre.metre_classifier import MetreClassifier, CompilationsSingleton
from rupo.util.vocabulary import Vocabulary
from rupo.accents.dict import AccentDict
from rupo.accents.classifier import MLAccentClassifier


class Generator(object):
    """
    Генератор стихов
    """
    def __init__(self, model_container, vocabulary: Vocabulary):
        """
        :param model_container: модель с методом get_model.
        :param vocabulary: словарь с индексами.
        """
        self.model_container = model_container
        self.vocabulary = vocabulary

    def generate_poem(self, metre_schema: str="-+", rhyme_pattern: str="abab", n_syllables: int=8,
                      letters_to_rhymes: dict=None) -> str:
        """
        Генерация стихотворения с выбранными параметрами.

        :param metre_schema: схема метра.
        :param rhyme_pattern: схема рифмы.
        :param n_syllables: количество слогов в строке.
        :param letters_to_rhymes: заданные рифмы.
        :return: стихотворение.
        """
        metre_pattern = ""
        while len(metre_pattern) <= n_syllables:
            metre_pattern += metre_schema
        metre_pattern = metre_pattern[:n_syllables]
        metre_filter = MetreFilter(metre_pattern)

        rhyme_filter = RhymeFilter(rhyme_pattern, letters_to_rhymes)

        prev_word_indices = []
        lines = []
        while rhyme_filter.position >= 0:
            words = self.generate_line(metre_filter, rhyme_filter, prev_word_indices)
            lines.append(" ".join(reversed(words)))
        return "\n".join(reversed(lines)) + "\n"

    def generate_line(self, metre_filter: MetreFilter, rhyme_filter: RhymeFilter,
                      prev_word_indices: List[int]) -> List[str]:
        """
        Генерация одной строки с заданными шаблонами метра и рифмы.

        :param metre_filter: фильтр по метру.
        :param rhyme_filter: фильтр по рифме.
        :param prev_word_indices: индексы предыдущих слов.
        :return: слова строка.
        """
        metre_filter.reset()
        result = []
        word_index = self.generate_word(prev_word_indices, [metre_filter, rhyme_filter])
        prev_word_indices.append(word_index)
        result.append(self.vocabulary.get_word(word_index).text.lower())
        while metre_filter.position >= 0:
            word_index = self.generate_word(prev_word_indices, [metre_filter])
            prev_word_indices.append(word_index)
            result.append(self.vocabulary.get_word(word_index).text.lower())
        return result

    def generate_word(self, prev_word_indices: List[int], filters: List[Filter]) -> int:
        """
        Генерация нового слова на основе предыдущих с учётом фильтров.

        :param prev_word_indices: индексы предыдущих слов.
        :param filters: фильтры модели.
        :return: индекс нового слова.
        """
        model = self.model_container.get_model(prev_word_indices)
        for f in filters:
            model = f.filter_model(model, self.vocabulary)
        if sum(model) == 0:
            model = self.model_container.get_model([])
            for f in filters:
                model = f.filter_model(model, self.vocabulary)
            if sum(model) == 0:
                model = self.model_container.get_model([])
        word_index = Generator.__choose(model)
        word = self.vocabulary.get_word(word_index)
        for f in filters:
            f.pass_word(word)
        return word_index

    def generate_poem_by_line(self, line: str, rhyme_pattern: str,
                              accent_dict: AccentDict, accents_classifier: MLAccentClassifier) -> str:
        """
        Генерация стихотвторения по одной строчке.

        :param accent_dict: словарь ударений.
        :param accents_classifier: классификатор.
        :param line: строчка.
        :param rhyme_pattern: шаблон рифмы.
        :return: стихотворение
        """
        markup, result = MetreClassifier.improve_markup(Phonetics.process_text(line, accent_dict), accents_classifier)
        rhyme_word = markup.lines[0].words[-1]
        count_syllables = sum([len(word.syllables) for word in markup.lines[0].words])
        metre_pattern = CompilationsSingleton.get().get_patterns(result.metre, count_syllables)[0]
        metre_pattern = metre_pattern.lower().replace("s", "+").replace("u", "-")
        letters_to_rhymes = {rhyme_pattern[0]: {rhyme_word}}
        generated = self.generate_poem(metre_pattern, rhyme_pattern, len(metre_pattern), letters_to_rhymes)
        poem = line + "\n" + "\n".join(generated.split("\n")[1:])
        return poem

    @staticmethod
    def __choose(model: np.array):
        """
        Выбор слова из языковой модели.

        :param model: языковая модель.
        :return: слово из модели.
        """
        return choice(range(len(model)), 1, p=model)[0]
